"""Defines a logger which logs the current training state."""

from pathlib import Path
from typing import Literal

from xax.task.logger import LoggerImpl, LogLine


class StateLogger(LoggerImpl):
    def __init__(
        self,
        run_directory: str | Path,
        flush_immediately: bool = False,
        open_mode: Literal["w", "a"] = "w",
        line_sep: str = "\n",
        remove_unicode_from_namespaces: bool = True,
    ) -> None:
        super().__init__(float("inf"))

        self.run_directory = Path(run_directory).expanduser().resolve()

        self.flush_immediately = flush_immediately
        self.open_mode = open_mode
        self.line_sep = line_sep
        self.remove_unicode_from_namespaces = remove_unicode_from_namespaces

    def log_file(self, name: str, contents: str) -> None:
        with open(self.run_directory / name, "w") as f:
            f.write(contents)

    def write(self, line: LogLine) -> None:
        pass



"""Defines a mixin for storing any task artifacts."""

import functools
import inspect
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Self, TypeVar

import jax

from xax.core.conf import field, get_run_dir
from xax.core.state import State
from xax.nn.parallel import is_master
from xax.task.base import BaseConfig, BaseTask
from xax.utils.experiments import stage_environment
from xax.utils.logging import LOG_STATUS
from xax.utils.text import show_info

logger = logging.getLogger(__name__)


@jax.tree_util.register_dataclass
@dataclass
class ArtifactsConfig(BaseConfig):
    exp_dir: str | None = field(None, help="The fixed experiment directory")


Config = TypeVar("Config", bound=ArtifactsConfig)


class ArtifactsMixin(BaseTask[Config]):
    _exp_dir: Path | None

    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self._exp_dir = None

    @functools.cached_property
    def run_dir(self) -> Path:
        run_dir = get_run_dir()
        if run_dir is None:
            task_file = inspect.getfile(self.__class__)
            run_dir = Path(task_file).resolve().parent
        return run_dir / self.task_name

    @property
    def exp_dir(self) -> Path:
        return self.get_exp_dir()

    def set_exp_dir(self, exp_dir: str | Path) -> Self:
        self._exp_dir = Path(exp_dir).expanduser().resolve()
        return self

    def get_exp_dir(self) -> Path:
        if self._exp_dir is not None:
            return self._exp_dir

        if self.config.exp_dir is not None:
            exp_dir = Path(self.config.exp_dir).expanduser().resolve()
            exp_dir.mkdir(parents=True, exist_ok=True)
            self._exp_dir = exp_dir
            logger.log(LOG_STATUS, self._exp_dir)
            return self._exp_dir

        def get_exp_dir(run_id: int) -> Path:
            return self.run_dir / f"run_{run_id}"

        run_id = 0
        while (exp_dir := get_exp_dir(run_id)).is_dir():
            run_id += 1
        exp_dir.mkdir(exist_ok=True, parents=True)
        self._exp_dir = exp_dir.expanduser().resolve()
        logger.log(LOG_STATUS, self._exp_dir)
        return self._exp_dir

    @functools.lru_cache(maxsize=None)
    def stage_environment(self) -> Path | None:
        stage_dir = (self.exp_dir / "code").resolve()
        try:
            stage_environment(self, stage_dir)
        except Exception:
            logger.exception("Failed to stage environment!")
            return None
        return stage_dir

    def on_training_end(self, state: State) -> State:
        state = super().on_training_end(state)

        if is_master():
            if self._exp_dir is None:
                show_info("Exiting training job", important=True)
            else:
                show_info(f"Exiting training job for {self.exp_dir}", important=True)

        return state



"""Normalization utilities."""

from typing import Literal, cast, get_args

import jax.numpy as jnp
from jaxtyping import Array

NormType = Literal["l1", "l2"]


def cast_norm_type(norm: str) -> NormType:
    if norm not in get_args(NormType):
        raise ValueError(f"Invalid norm: {norm}")
    return cast(NormType, norm)


def get_norm(x: Array, norm: NormType) -> Array:
    match norm:
        case "l1":
            return jnp.abs(x)
        case "l2":
            return jnp.square(x)
        case _:
            raise ValueError(f"Invalid norm: {norm}")



"""Defines a mixin for handling model checkpointing."""

import io
import json
import logging
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import Generic, Literal, TypeVar, cast, overload

import equinox as eqx
import jax
import optax
from jaxtyping import PyTree
from omegaconf import DictConfig, OmegaConf

from xax.core.conf import field
from xax.core.state import State
from xax.nn.parallel import is_master
from xax.task.mixins.artifacts import ArtifactsConfig, ArtifactsMixin

logger = logging.getLogger(__name__)

CheckpointPart = Literal["model", "opt", "opt_state", "state", "config", "model_state_config", "all"]


def get_ckpt_path(exp_dir: Path, state: State | None = None) -> Path:
    """Defines the path to the checkpoint for a given state.

    Args:
        exp_dir: The experiment directory
        state: The current trainer state

    Returns:
        The path to the checkpoint file.
    """
    if state is None:
        return exp_dir / "checkpoints" / "ckpt.bin"
    return exp_dir / "checkpoints" / f"ckpt.{state.num_steps}.bin"


@jax.tree_util.register_dataclass
@dataclass
class CheckpointingConfig(ArtifactsConfig):
    save_every_n_steps: int | None = field(None, help="Save a checkpoint every N steps")
    save_every_n_seconds: float | None = field(60.0 * 60.0, help="Save a checkpoint every N seconds")
    only_save_most_recent: bool = field(True, help="Only keep the most recent checkpoint")
    load_from_ckpt_path: str | None = field(None, help="If set, load initial model weights from this path")
    load_ckpt_strict: bool = field(True, help="If set, only load weights for which have a matching key in the model")


Config = TypeVar("Config", bound=CheckpointingConfig)


class CheckpointingMixin(ArtifactsMixin[Config], Generic[Config]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

        self.__last_ckpt_time = 0.0

    def get_ckpt_path(self, state: State | None = None) -> Path:
        return get_ckpt_path(self.exp_dir, state)

    def get_init_ckpt_path(self) -> Path | None:
        if self._exp_dir is not None:
            if (ckpt_path := self.get_ckpt_path()).exists():
                return ckpt_path
        if self.config.load_from_ckpt_path is not None:
            ckpt_path = Path(self.config.load_from_ckpt_path)
            assert ckpt_path.exists(), f"Checkpoint path {ckpt_path} does not exist."
            return ckpt_path
        return None

    def should_checkpoint(self, state: State) -> bool:
        if self.config.save_every_n_steps is not None:
            if state.num_steps % self.config.save_every_n_steps == 0:
                return True
        if self.config.save_every_n_seconds is not None:
            last_time, cur_time = self.__last_ckpt_time, state.elapsed_time_s.item()
            if cur_time - last_time >= self.config.save_every_n_seconds:
                self.__last_ckpt_time = cur_time
                return True
        return False

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["all"],
        model_template: PyTree,
        optimizer_template: PyTree,
        opt_state_template: PyTree,
    ) -> tuple[PyTree, optax.GradientTransformation, optax.OptState, State, Config]: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["model_state_config"],
        model_template: PyTree,
    ) -> tuple[PyTree, State, Config]: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["model"],
        model_template: PyTree,
    ) -> PyTree: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["opt"],
        optimizer_template: PyTree,
    ) -> optax.GradientTransformation: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["opt_state"],
        opt_state_template: PyTree,
    ) -> optax.OptState: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["state"],
    ) -> State: ...

    @overload
    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: Literal["config"],
    ) -> Config: ...

    def load_ckpt_with_template(
        self,
        path: Path,
        *,
        part: CheckpointPart = "all",
        model_template: PyTree | None = None,
        optimizer_template: PyTree | None = None,
        opt_state_template: PyTree | None = None,
    ) -> (
        tuple[PyTree, optax.GradientTransformation, optax.OptState, State, Config]
        | tuple[PyTree, State, Config]
        | PyTree
        | optax.GradientTransformation
        | optax.OptState
        | State
        | Config
    ):
        """Load a checkpoint.

        Args:
            path: Path to the checkpoint directory
            part: Which part of the checkpoint to load
            model_template: Template model with correct structure but uninitialized weights
            optimizer_template: Template optimizer with correct structure but uninitialized weights
            opt_state_template: Template optimizer state with correct structure but uninitialized weights

        Returns:
            The requested checkpoint components
        """
        with tarfile.open(path, "r:gz") as tar:

            def get_model() -> PyTree:
                if model_template is None:
                    raise ValueError("model_template must be provided to load model weights")
                if (model := tar.extractfile("model")) is None:
                    raise ValueError(f"Checkpoint does not contain a model file: {path}")
                return eqx.tree_deserialise_leaves(io.BytesIO(model.read()), model_template)

            def get_opt() -> optax.GradientTransformation:
                if optimizer_template is None:
                    raise ValueError("optimizer_template must be provided to load optimizer")
                if (opt := tar.extractfile("optimizer")) is None:
                    raise ValueError(f"Checkpoint does not contain an optimizer file: {path}")
                return eqx.tree_deserialise_leaves(io.BytesIO(opt.read()), optimizer_template)

            def get_opt_state() -> optax.OptState:
                if opt_state_template is None:
                    raise ValueError("opt_state_template must be provided to load optimizer state")
                if (opt_state := tar.extractfile("opt_state")) is None:
                    raise ValueError(f"Checkpoint does not contain an optimizer state file: {path}")
                return eqx.tree_deserialise_leaves(io.BytesIO(opt_state.read()), opt_state_template)

            def get_state() -> State:
                if (state := tar.extractfile("state")) is None:
                    raise ValueError(f"Checkpoint does not contain a state file: {path}")
                return State.from_dict(**json.loads(state.read().decode()))

            def get_config() -> Config:
                if (config := tar.extractfile("config")) is None:
                    raise ValueError(f"Checkpoint does not contain a config file: {path}")
                return self.get_config(cast(DictConfig, OmegaConf.load(config)), use_cli=False)

            match part:
                case "model":
                    return get_model()
                case "opt":
                    return get_opt()
                case "opt_state":
                    return get_opt_state()
                case "state":
                    return get_state()
                case "config":
                    return get_config()
                case "model_state_config":
                    return get_model(), get_state(), get_config()
                case "all":
                    return get_model(), get_opt(), get_opt_state(), get_state(), get_config()
                case _:
                    raise ValueError(f"Invalid checkpoint part: {part}")

    def save_checkpoint(
        self,
        model: PyTree | None = None,
        optimizer: optax.GradientTransformation | None = None,
        opt_state: optax.OptState | None = None,
        aux_data: PyTree | None = None,
        state: State | None = None,
    ) -> Path:
        """Save a checkpoint.

        Args:
            model: The model to save
            state: The current training state
            optimizer: The optimizer to save
            aux_data: Additional data to save
            opt_state: The optimizer state to save

        Returns:
            Path to the saved checkpoint
        """
        ckpt_path = self.get_ckpt_path(state)

        if not is_master():
            return ckpt_path

        # Gets the path to the last checkpoint
        logger.info("Saving checkpoint to %s", ckpt_path)
        last_ckpt_path = self.get_ckpt_path()
        ckpt_path.parent.mkdir(exist_ok=True, parents=True)

        # Potentially removes the last checkpoint
        if last_ckpt_path.exists() and self.config.only_save_most_recent:
            if (base_ckpt := last_ckpt_path.resolve()).is_file():
                base_ckpt.unlink()

        # Save the checkpoint components
        with tarfile.open(ckpt_path, "w:gz") as tar:

            def add_file(name: str, buf: io.BytesIO) -> None:
                tarinfo = tarfile.TarInfo(name)
                tarinfo.size = buf.tell()
                buf.seek(0)
                tar.addfile(tarinfo, buf)

            # Save model using Equinox
            if model is not None:
                with io.BytesIO() as buf:
                    eqx.tree_serialise_leaves(buf, model)
                    add_file("model", buf)

            # Save optimizer using Equinox
            if optimizer is not None:
                with io.BytesIO() as buf:
                    eqx.tree_serialise_leaves(buf, optimizer)
                    add_file("optimizer", buf)

            # Save optimizer state using Equinox
            if opt_state is not None:
                with io.BytesIO() as buf:
                    eqx.tree_serialise_leaves(buf, opt_state)
                    add_file("opt_state", buf)

            # Save aux data using Equinox.
            if aux_data is not None:
                with io.BytesIO() as buf:
                    eqx.tree_serialise_leaves(buf, aux_data)
                    add_file("aux_data", buf)

            # Save state and config as JSON
            def add_file_bytes(name: str, data: bytes) -> None:  # noqa: ANN401
                info = tarfile.TarInfo(name=name)
                info.size = len(data)
                tar.addfile(info, io.BytesIO(data))

            if state is not None:
                add_file_bytes("state", json.dumps(state.to_dict(), indent=2).encode())
            add_file_bytes("config", OmegaConf.to_yaml(self.config).encode())

        # Updates the symlink to the new checkpoint
        last_ckpt_path.unlink(missing_ok=True)
        try:
            last_ckpt_path.symlink_to(ckpt_path.relative_to(last_ckpt_path.parent))
        except FileExistsError:
            logger.exception("Exception while trying to update %s", ckpt_path)

        # Calls the base callback
        self.on_after_checkpoint_save(ckpt_path, state)

        return ckpt_path



"""Defines a logger that logs to stdout."""

import datetime
import logging
import sys
from collections import deque
from typing import Any, Deque, Mapping, TextIO

from jaxtyping import Array

from xax.task.logger import (
    LogError,
    LogErrorSummary,
    LoggerImpl,
    LogLine,
    LogPing,
    LogScalar,
    LogStatus,
    LogString,
)
from xax.utils.text import Color, colored, format_timedelta


def format_number(value: int | float, precision: int) -> str:
    if isinstance(value, int):
        return f"{value:,}"  # Add commas to the number
    return f"{value:.{precision}g}"


def as_str(value: Any, precision: int) -> str:  # noqa: ANN401
    if isinstance(value, str):
        return f'"{value}"'
    if isinstance(value, Array):
        value = value.item()
    if isinstance(value, (int, float)):
        return format_number(value, precision)
    raise TypeError(f"Unexpected log type: {type(value)}")


class StdoutLogger(LoggerImpl):
    def __init__(
        self,
        write_fp: TextIO = sys.stdout,
        precision: int = 4,
        log_timers: bool = True,
        log_perf: bool = False,
        log_optim: bool = False,
        log_fp: bool = False,
        log_interval_seconds: float = 1.0,
        remove_temporary_after: datetime.timedelta = datetime.timedelta(seconds=10),
    ) -> None:
        """Defines a logger which shows a pop-up using Curses.

        Args:
            write_fp: The file to write logs to.
            precision: The integer precision to use when logging scalars.
            log_timers: Whether to log timers.
            log_perf: Whether to log performance metrics.
            log_optim: Whether to log optimizer parameters.
            log_fp: Whether to log floating point parameters.
            log_interval_seconds: The interval between successive log lines.
            remove_temporary_after: The time after which temporary toasts
                are removed.
        """
        super().__init__(log_interval_seconds)

        self.write_fp = write_fp
        self.log_timers = log_timers
        self.log_perf = log_perf
        self.log_fp = log_fp
        self.log_optim = log_optim
        self.precision = precision
        self.remove_temporary_after = remove_temporary_after
        self.logger = logging.getLogger("stdout")

        self.statuses: Deque[tuple[str, datetime.datetime]] = deque()
        self.pings: Deque[tuple[str, datetime.datetime]] = deque()
        self.errors: Deque[tuple[str, datetime.datetime]] = deque()
        self.error_summary: tuple[str, datetime.datetime] | None = None

    def start(self) -> None:
        return super().start()

    def stop(self) -> None:
        self.write_queues()
        return super().stop()

    def write_separator(self) -> None:
        self.write_fp.write("\033[2J\033[H")

    def write_state_window(self, line: LogLine) -> None:
        state_info: dict[str, str] = {
            "Steps": format_number(int(line.state.num_steps.item()), 0),
            "Samples": format_number(int(line.state.num_samples.item()), 0),
            "Elapsed Time": format_timedelta(datetime.timedelta(seconds=line.state.elapsed_time_s.item()), short=True),
        }

        colored_prefix = colored("Phase: ", "grey", bold=True)
        colored_phase = colored(line.state.phase, "green" if line.state.phase == "train" else "yellow", bold=True)
        self.write_fp.write(f"{colored_prefix}{colored_phase}\n")
        for k, v in state_info.items():
            self.write_fp.write(f" ↪ {k}: {colored(v, 'cyan')}\n")

    def write_log_window(self, line: LogLine) -> None:
        namespace_to_lines: dict[str, dict[str, str]] = {}

        def add_logs(
            log: Mapping[str, Mapping[str, LogScalar | LogString]],
            namespace_to_lines: dict[str, dict[str, str]],
        ) -> None:
            for namespace, values in log.items():
                for k, v in values.items():
                    if v.secondary:
                        continue
                    if namespace not in namespace_to_lines:
                        namespace_to_lines[namespace] = {}
                    v_str = as_str(v.value, self.precision)
                    namespace_to_lines[namespace][k] = v_str

        add_logs(line.scalars, namespace_to_lines)
        add_logs(line.strings, namespace_to_lines)
        if not namespace_to_lines:
            return

        for namespace, lines in sorted(namespace_to_lines.items()):
            self.write_fp.write(f"\n{colored(namespace, 'cyan', bold=True)}\n")
            for k, v in lines.items():
                self.write_fp.write(f" ↪ {k}: {v}\n")

    def write_queue(self, title: str, q: Deque[tuple[str, datetime.datetime]], remove: bool, color: Color) -> None:
        if not q:
            return

        self.write_fp.write(f"\n{colored(title, 'grey', bold=True)}\n")
        self.write_fp.write("\n".join(f" ✦ {colored(msg, color)}" for msg, _ in reversed(q)))
        self.write_fp.write("\n")

        if remove:
            now = datetime.datetime.now()
            while q and now - q[0][1] > self.remove_temporary_after:
                q.popleft()

    def write_queues(self) -> None:
        self.write_queue("Status", self.statuses, False, "green")
        self.write_queue("Pings", self.pings, True, "cyan")
        self.write_queue("Errors", self.errors, False, "red")

    def write_error_summary_to_screen(self) -> None:
        if self.error_summary is not None:
            summary, timestamp = self.error_summary
            timestamp_string = timestamp.strftime("%Y-%m-%d %H:%M:%S")
            self.write_fp.write(f"\n{colored('Exception summary', 'grey', bold=True)}")
            self.write_fp.write(f" {colored(timestamp_string, 'grey')}")
            self.write_fp.write(f"\n{summary}")

    def write(self, line: LogLine) -> None:
        self.write_separator()
        self.write_state_window(line)
        self.write_log_window(line)
        self.write_queues()
        self.write_error_summary_to_screen()
        sys.stdout.flush()

    def write_error_summary(self, error_summary: LogErrorSummary) -> None:
        self.error_summary = error_summary.message, datetime.datetime.now()

    def write_error(self, error: LogError) -> None:
        self.errors.append((error.message_with_location, datetime.datetime.now()))

    def write_status(self, status: LogStatus) -> None:
        self.statuses.append((status.message, datetime.datetime.now()))

    def write_ping(self, ping: LogPing) -> None:
        self.pings.append((ping.message, datetime.datetime.now()))



"""Defines some useful Jax debugging utilities."""

from collections import deque
from collections.abc import Iterable, Mapping
from typing import Any, Callable, Deque

import jax
import jax.numpy as jnp
from jaxtyping import Array


def get_named_leaves(
    obj: Any,  # noqa: ANN401
    is_leaf: Callable[[Any], bool] = lambda x: isinstance(x, Array),  # noqa: ANN401
    max_depth: int = 100,
) -> list[tuple[str, Any]]:  # noqa: ANN401
    ret: list[tuple[str, Any]] = []
    q: Deque[tuple[int, str, Any]] = deque()  # noqa: ANN401
    q.append((0, "", obj))

    while q:
        depth, name, node = q.popleft()

        if depth > max_depth:
            continue

        if hasattr(node, "__dict__") and isinstance(node.__dict__, Mapping):
            for cname, cnode in node.__dict__.items():
                gname = f"{name}.{cname}" if name else cname
                if is_leaf(cnode):
                    ret.append((gname, cnode))
                else:
                    q.append((depth + 1, gname, cnode))

        elif isinstance(node, Mapping):
            for cname, cnode in node.items():
                gname = f"{name}.{cname}" if name else cname
                if is_leaf(cnode):
                    ret.append((gname, cnode))
                else:
                    q.append((depth + 1, gname, cnode))

        elif isinstance(node, Iterable):
            for i, cnode in enumerate(node):
                gname = f"{name}.{i}" if name else str(i)
                if is_leaf(cnode):
                    ret.append((gname, cnode))
                else:
                    q.append((depth + 1, gname, cnode))

    return ret


def breakpoint_if_nan(x: Array) -> None:
    jax.lax.cond(jnp.any(jnp.isnan(x)), lambda: jax.debug.breakpoint(), lambda: None)


def log_if_nan(x: Array, loc: str) -> None:
    jax.lax.cond(jnp.any(jnp.isnan(x)), lambda: jax.debug.print("=== NaNs: {loc} ===", loc=loc), lambda: None)



"""Defines the core logger.

A common problem when quickly prototyping ML models is nicely logging images,
videos, audio, or other data. Additionally, logging on every step can be
overwhelming. This logger implements a number of convenience functions to
take heterogeneous input data and put it into a standard format, which can
then be used by downstream loggers to actually log the data. For example, this
logger will automatically tile multiple images into a single image, add
captions to images, and so on.
"""

import functools
import logging
import math
import re
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from types import TracebackType
from typing import (
    Any,
    Callable,
    Iterator,
    Literal,
    Self,
    Sequence,
    TypeVar,
    cast,
    get_args,
)

import jax
import jax.numpy as jnp
import numpy as np
from jax._src.core import ClosedJaxpr
from jaxtyping import Array
from PIL import Image, ImageDraw, ImageFont
from PIL.Image import Image as PILImage

from xax.core.state import Phase, State
from xax.utils.experiments import ContextTimer, IntervalTicker
from xax.utils.logging import LOG_ERROR_SUMMARY, LOG_PING, LOG_STATUS

logger = logging.getLogger(__name__)

T = TypeVar("T")
LogT = TypeVar("LogT")
Number = int | float | Array | np.ndarray

ChannelSelectMode = Literal["first", "last", "mean"]

DEFAULT_NAMESPACE = "value"

NAMESPACE_STACK: list[str] = []


def standardize_text(text: str, max_line_length: int | None = None, remove_non_ascii: bool = False) -> list[str]:
    """Standardizes a text string to a list of lines.

    Args:
        text: The text to standardize
        max_line_length: If set, truncate lines to this length
        remove_non_ascii: Remove non-ASCII characters if present

    Returns:
        The standardized text lines
    """

    def _chunk_lines(text: str, max_length: int) -> Iterator[str]:
        for i in range(0, len(text), max_length):
            yield text[i : i + max_length]

    if remove_non_ascii:
        text = "".join(char for char in text if ord(char) < 128)
    lines = [re.sub(r"\s+", " ", line) for line in re.split(r"[\n\r]+", text.strip())]
    if max_line_length is not None:
        lines = [subline for line in lines for subline in _chunk_lines(line, max_line_length)]
    return lines


def make_human_viewable_resolution(
    image: PILImage,
    interpolation: Image.Resampling = Image.Resampling.LANCZOS,
    trg_res: tuple[int, int] = (512, 512),
) -> PILImage:
    """Resizes image to human-viewable resolution.

    Args:
        image: The image to resize, with shape (C, H, W)
        interpolation: Interpolation mode to use for image resizing
        trg_res: The target image resolution; the image will be reshaped to
            have approximately the same area as an image with this resolution

    Returns:
        The resized image
    """
    width, height = image.size
    trg_height, trg_width = trg_res
    factor = math.sqrt((trg_height * trg_width) / (height * width))
    new_height, new_width = int(height * factor), int(width * factor)
    return image.resize((new_width, new_height), interpolation)


class namespace_context:  # noqa: N801
    def __init__(self, name: str | None) -> None:
        self._name = name
        self._prev_stack: list[str] | None = None

    def __enter__(self) -> None:
        if self._name is None:
            self._prev_stack = NAMESPACE_STACK[:]
            NAMESPACE_STACK.clear()
        else:
            NAMESPACE_STACK.append(self._name)

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        if self._prev_stack is not None:
            NAMESPACE_STACK[:] = self._prev_stack
        else:
            NAMESPACE_STACK.pop()


def normalize(x: np.ndarray) -> np.ndarray:
    return (x - x.min()) / (x.max() - x.min())


def ternary_search_optimal_side_counts(height: int, width: int, count: int) -> tuple[int, int]:
    min_factors = [i for i in range(1, math.ceil(math.sqrt(count)) + 1) if count % i == 0]
    max_factors = [i for i in min_factors[::-1] if i * i != count]
    factors = [(i, count // i) for i in min_factors] + [(count // i, i) for i in max_factors]

    lo, hi = 0, len(factors) - 1

    def penalty(i: int) -> float:
        hval, wval = factors[i]
        h, w = hval * height, wval * width
        return -(min(h, w) ** 2)

    # Runs ternary search to minimize penalty.
    while lo < hi - 2:
        lmid, rmid = (lo * 2 + hi) // 3, (lo + hi * 2) // 3
        if penalty(lmid) > penalty(rmid):
            lo = lmid
        else:
            hi = rmid

    # Returns the lowest-penalty configuration.
    mid = (lo + hi) // 2
    plo, pmid, phi = penalty(lo), penalty(mid), penalty(hi)

    if pmid <= plo and pmid <= phi:
        return factors[mid]
    elif plo <= phi:
        return factors[lo]
    else:
        return factors[hi]


def tile_images_different_sizes(images: list[PILImage], sep: int) -> PILImage:
    """Tiles a list of images into a single image, even if they have different sizes.

    Args:
        images: The images to tile.
        sep: The separation between adjacent images.

    Returns:
        The tiled image.
    """
    total_width, max_height = sum(image.width for image in images), max(image.height for image in images)
    tiled = Image.new("RGB", (total_width + (len(images) - 1) * sep, max_height))
    x = 0
    for image in images:
        tiled.paste(image, (x, 0))
        x += image.width + sep
    return tiled


def tile_images(images: list[PILImage], sep: int = 0) -> PILImage:
    """Tiles a list of images into a single image.

    Args:
        images: The images to tile.
        sep: The separation between adjacent images.

    Returns:
        The tiled image.
    """
    if not images:
        return Image.new("RGB", (0, 0))

    # Gets the optimal side counts.
    height, width = images[0].height, images[0].width
    if not all(image.size == images[0].size for image in images):
        return tile_images_different_sizes(images, sep)

    hside, wside = ternary_search_optimal_side_counts(height, width, len(images))

    # Tiles the images.
    tiled = Image.new("RGB", (wside * width + (wside - 1) * sep, hside * height + (hside - 1) * sep))
    for i, image in enumerate(images):
        x, y = i % wside, i // wside
        tiled.paste(image, (x * (width + sep), y * (height + sep)))

    return tiled


def as_numpy(array: Array | np.ndarray) -> np.ndarray:
    """Convert a JAX array or numpy array to numpy array."""
    if isinstance(array, np.ndarray):
        return array
    array = jax.device_get(array)
    if jax.dtypes.issubdtype(array.dtype, jnp.floating):
        array = array.astype(jnp.float32)
    elif jax.dtypes.issubdtype(array.dtype, jnp.integer):
        array = array.astype(jnp.int32)
    elif jax.dtypes.issubdtype(array.dtype, jnp.bool_):
        array = array.astype(jnp.bool_)
    return np.array(array)


def as_numpy_opt(array: Array | np.ndarray | None) -> np.ndarray | None:
    """Convert an optional JAX array or numpy array to numpy array."""
    if array is None:
        return None
    return as_numpy(array)


@dataclass(kw_only=True)
class LogString:
    value: str
    secondary: bool


@dataclass(kw_only=True)
class LogImage:
    image: PILImage


@dataclass(kw_only=True)
class LogVideo:
    """Container for video data and metadata.

    Attributes:
        frames: Video frames as a numpy array of shape (T,H,W,C)
        fps: Frames per second
    """

    frames: np.ndarray
    fps: int


@dataclass(kw_only=True)
class LogScalar:
    value: Number
    secondary: bool


@dataclass(kw_only=True)
class LogDistribution:
    mean: Number
    std: Number


@dataclass(kw_only=True)
class LogHistogram:
    min: Number
    max: Number
    num: int
    sum: Number
    sum_squares: Number
    bucket_limits: list[float]
    bucket_counts: list[int]


@dataclass(kw_only=True)
class LogMesh:
    vertices: np.ndarray
    colors: np.ndarray | None
    faces: np.ndarray | None
    config_dict: dict[str, Any] | None  # noqa: ANN401


@dataclass(kw_only=True)
class LogGraph:
    computation: ClosedJaxpr


@dataclass(kw_only=True)
class LogLine:
    state: State
    scalars: dict[str, dict[str, LogScalar]]
    distributions: dict[str, dict[str, LogDistribution]]
    histograms: dict[str, dict[str, LogHistogram]]
    strings: dict[str, dict[str, LogString]]
    images: dict[str, dict[str, LogImage]]
    videos: dict[str, dict[str, LogVideo]]
    meshes: dict[str, dict[str, LogMesh]]


@dataclass(kw_only=True)
class LogErrorSummary:
    message: str


@dataclass(kw_only=True)
class LogError:
    message: str
    location: str | None = None

    @property
    def message_with_location(self) -> str:
        message = self.message
        if self.location is not None:
            message += f" ({self.location})"
        return message


@dataclass(kw_only=True)
class LogStatus:
    message: str
    created: float
    filename: str | None = None
    lineno: int | None = None


@dataclass(kw_only=True)
class LogPing:
    message: str
    created: float
    filename: str | None = None
    lineno: int | None = None


def get_image(image: np.ndarray | Array | PILImage, target_resolution: tuple[int, int] | None = None) -> LogImage:
    if not isinstance(image, (np.ndarray, Array, PILImage)):
        raise ValueError(f"Unsupported image type: {type(image)}")
    if isinstance(image, Array):
        image = as_numpy(image)
    if isinstance(image, np.ndarray):
        if image.ndim == 2:
            image = np.expand_dims(image, axis=-1)
        if image.ndim != 3:
            raise RuntimeError(f"Expected image to have shape HW, HWC, or CHW, got {image.shape}")

        # Normalizes the image and converts to integer.
        if np.issubdtype(image.dtype, np.floating):
            image = (normalize(image) * 255).round().astype(np.uint8)
        elif image.dtype == np.uint8:
            pass
        else:
            raise ValueError(f"Unsupported image dtype: {image.dtype}")

        # Converts to a PIL image.
        if image.shape[-1] == 1:
            image = Image.fromarray(image[..., 0])
        elif image.shape[-1] == 3:
            image = Image.fromarray(image)
        elif image.shape[0] == 1:
            image = Image.fromarray(image[0])
        elif image.shape[0] == 3:
            image = Image.fromarray(image.transpose(1, 2, 0))
        else:
            raise ValueError(f"Unsupported image shape: {image.shape}")

    if target_resolution is not None:
        image = make_human_viewable_resolution(image, trg_res=target_resolution)
    return LogImage(image=image)


def image_with_text(
    image: PILImage,
    text: list[str],
    max_num_lines: int | None,
    line_spacing: int,
    centered: bool,
) -> LogImage:
    """Adds a text label to an image.

    Args:
        image: The image to label, with shape (C, H, W)
        text: The text label for the image
        max_num_lines: The number of lines of spacing to add to the bottom
            of the image
        line_spacing: The spacing between adjacent lines
        centered: If set, center the text labels, otherwise align to the left

    Returns:
        The image with a text label
    """
    if not text:
        return LogImage(image=image)
    if max_num_lines is None:
        max_num_lines = len(text)
    else:
        text = text[:max_num_lines]
    width, height = image.size
    font: ImageFont.ImageFont | ImageFont.FreeTypeFont = ImageFont.load_default()
    _, _, _, line_height = font.getbbox(text[0])
    new_width, new_height = width, int(height + line_spacing + max_num_lines * (line_height + line_spacing))
    padded_image = Image.new(image.mode, (new_width, new_height), 255)
    padded_image.paste(image, (0, 0))
    drawer = ImageDraw.Draw(padded_image)
    for i, text_line in enumerate(text):
        text_line_top = height + line_spacing + i * (line_height + line_spacing)
        if centered:
            _, _, line_width, _ = font.getbbox(text_line)
            text_line_left = (width - line_width) / 2
            drawer.text((text_line_left, text_line_top), text_line, font=font, fill=0)
        else:
            drawer.text((line_spacing, text_line_top), text_line, font=font, fill=0)
    return LogImage(image=padded_image)


def get_video(video: np.ndarray | Array, fps: int = 30) -> LogVideo:
    """Converts video data to standard format.

    Args:
        video: The video frames. Can be:
            - A numpy array of shape (T, H, W, C) or (T, C, H, W)
            - A JAX array of shape (T, H, W, C) or (T, C, H, W)
        fps: Frames per second

    Returns:
        LogVideo containing standardized video frames
    """
    if isinstance(video, Array):
        video = as_numpy(video)

    if not isinstance(video, np.ndarray):
        raise ValueError(f"Unsupported video type: {type(video)}")

    # Handle different dimension orderings
    if video.ndim != 4:
        raise ValueError(f"Expected video array of shape (T, H, W, C) or (T, C, H, W), got shape {video.shape}")

    if video.shape[1] == 3:  # (T,C,H,W) format
        video = video.transpose(0, 2, 3, 1)

    # Normalize and convert to uint8 if needed
    if np.issubdtype(video.dtype, np.floating):
        video = (normalize(video) * 255).round().astype(np.uint8)
    elif video.dtype != np.uint8:
        raise ValueError(f"Unsupported video dtype: {video.dtype}")

    return LogVideo(frames=video, fps=fps)


class LoggerImpl(ABC):
    def __init__(self, log_interval_seconds: float = 1.0) -> None:
        """Defines some default behavior for loggers.

        Every logger needs to implement the ``write`` function, which handles
        actually writing the logs to wherever they needs to go. The basic
        class implements a simple interval-based logging scheme to avoid
        writing too many log lines.

        Args:
            log_interval_seconds: The interval between successive log lines.
        """
        super().__init__()

        self.tickers = {phase: IntervalTicker(log_interval_seconds) for phase in get_args(Phase)}

    def start(self) -> None:
        pass

    def stop(self) -> None:
        pass

    @abstractmethod
    def write(self, line: LogLine) -> None:
        """Handles writing the current log line.

        Args:
            line: The line to write.
        """

    def write_error_summary(self, error_summary: LogErrorSummary) -> None:
        """Handles writing an error summary.

        Args:
            error_summary: The error summary to write.
        """

    def write_error(self, error: LogError) -> None:
        """Handles writing an error line.

        Args:
            error: The error information to write.
        """

    def write_status(self, status: LogStatus) -> None:
        """Handles writing a status line.

        Args:
            status: The status to write.
        """

    def write_ping(self, ping: LogPing) -> None:
        """Handles writing a ping line.

        Args:
            ping: The ping to write.
        """

    def log_file(self, name: str, contents: str) -> None:
        """Logs a large text file.

        Args:
            name: The name of the file.
            contents: The contents of the file.
        """

    def should_log(self, state: State) -> bool:
        """Function that determines if the logger should log the current step.

        Args:
            state: The current step's state.

        Returns:
            If the logger should log the current step.
        """
        return self.tickers[state.phase].tick(state.elapsed_time_s.item())


class ToastHandler(logging.Handler):
    def __init__(self, logger: "Logger") -> None:
        super().__init__()

        self.logger = logger

    def emit(self, record: logging.LogRecord) -> None:
        try:
            if record.levelno == LOG_ERROR_SUMMARY:
                self.logger.write_error_summary(record.getMessage())
            elif record.levelno == LOG_STATUS:
                self.logger.write_status(record.getMessage(), record.filename, record.lineno)
            elif record.levelno in (LOG_PING, logging.WARNING):
                self.logger.write_ping(record.getMessage(), record.filename, record.lineno)
            elif record.levelno in (logging.ERROR, logging.CRITICAL, logging.WARNING):
                self.logger.write_error(record.getMessage(), f"{record.filename}:{record.lineno}")
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)

    def add_for_logger(self, logger: logging.Logger) -> None:
        # Removes existing ToastHandler.
        handlers_to_remove = []
        for handler in logger.handlers:
            if isinstance(handler, ToastHandler):
                handlers_to_remove.append(handler)
        for handler in handlers_to_remove:
            logger.removeHandler(handler)

        # Adds the new ToastHandler.
        logger.addHandler(self)


class Logger:
    """Defines an intermediate container which holds values to log somewhere else."""

    def __init__(self, default_namespace: str = DEFAULT_NAMESPACE) -> None:
        self.scalars: dict[str, dict[str, Callable[[], LogScalar]]] = defaultdict(dict)
        self.distributions: dict[str, dict[str, Callable[[], LogDistribution]]] = defaultdict(dict)
        self.histograms: dict[str, dict[str, Callable[[], LogHistogram]]] = defaultdict(dict)
        self.strings: dict[str, dict[str, Callable[[], LogString]]] = defaultdict(dict)
        self.images: dict[str, dict[str, Callable[[], LogImage]]] = defaultdict(dict)
        self.videos: dict[str, dict[str, Callable[[], LogVideo]]] = defaultdict(dict)
        self.meshes: dict[str, dict[str, Callable[[], LogMesh]]] = defaultdict(dict)
        self.default_namespace = default_namespace
        self.loggers: list[LoggerImpl] = []

        # Registers a logging handler to route log messages to the logger.
        root_logger = logging.getLogger()
        ToastHandler(self).add_for_logger(root_logger)

        # Flag when the logger is active.
        self.active = False

    def add_logger(self, *logger: LoggerImpl) -> None:
        """Add the logger, so that it gets called when `write` is called.

        Args:
            logger: The logger to add.
        """
        self.loggers.extend(logger)

    def pack(self, state: State) -> LogLine:
        return LogLine(
            state=state,
            scalars={k: {kk: v() for kk, v in v.items()} for k, v in self.scalars.items()},
            distributions={k: {kk: v() for kk, v in v.items()} for k, v in self.distributions.items()},
            histograms={k: {kk: v() for kk, v in v.items()} for k, v in self.histograms.items()},
            strings={k: {kk: v() for kk, v in v.items()} for k, v in self.strings.items()},
            images={k: {kk: v() for kk, v in v.items()} for k, v in self.images.items()},
            videos={k: {kk: v() for kk, v in v.items()} for k, v in self.videos.items()},
            meshes={k: {kk: v() for kk, v in v.items()} for k, v in self.meshes.items()},
        )

    def clear(self) -> None:
        self.scalars.clear()
        self.distributions.clear()
        self.histograms.clear()
        self.strings.clear()
        self.images.clear()
        self.videos.clear()
        self.meshes.clear()

    def write(self, state: State) -> None:
        """Writes the current step's logging information.

        Args:
            state: The current step's state.
        """
        should_log = [lg.should_log(state) for lg in self.loggers]
        if not any(should_log):
            self.clear()
            return
        line = self.pack(state)
        self.clear()
        for lg in (lg for lg, should_log in zip(self.loggers, should_log) if should_log):
            lg.write(line)

    def write_error_summary(self, error_summary: str) -> None:
        for logger in self.loggers:
            logger.write_error_summary(LogErrorSummary(message=error_summary))

    def write_error(self, message: str, location: str | None = None) -> None:
        for logger in self.loggers:
            logger.write_error(LogError(message=message, location=location))

    def write_status(
        self,
        message: str,
        filename: str | None = None,
        lineno: int | None = None,
        created: float | None = None,
    ) -> None:
        status = LogStatus(
            message=message,
            created=time.time() if created is None else created,
            filename=filename,
            lineno=lineno,
        )
        for logger in self.loggers:
            logger.write_status(status)

    def write_ping(
        self,
        message: str,
        filename: str | None = None,
        lineno: int | None = None,
        created: float | None = None,
    ) -> None:
        ping = LogPing(
            message=message,
            created=time.time() if created is None else created,
            filename=filename,
            lineno=lineno,
        )
        for logger in self.loggers:
            logger.write_ping(ping)

    def resolve_namespace(self, namespace: str | None = None) -> str:
        return "_".join([self.default_namespace if namespace is None else namespace] + NAMESPACE_STACK)

    def log_scalar(
        self,
        key: str,
        value: Callable[[], Number] | Number,
        *,
        namespace: str | None = None,
        secondary: bool = False,
    ) -> None:
        """Logs a scalar value.

        Args:
            key: The key being logged
            value: The scalar value being logged
            namespace: An optional logging namespace
            secondary: If set, treat this as a secondary value (meaning, it is
                less important than other values, and some downstream loggers
                will not display it)
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        if isinstance(value, jnp.ndarray):
            assert value.ndim == 0, f"Scalar must be a 0D array, got shape {value.shape}"

        @functools.lru_cache(maxsize=None)
        def scalar_future() -> LogScalar:
            with ContextTimer() as timer:
                value_concrete = value() if callable(value) else value
            logger.debug("Scalar Key: %s, Time: %s", key, timer.elapsed_time)
            return LogScalar(value=value_concrete, secondary=secondary)

        self.scalars[namespace][key] = scalar_future

    def log_distribution(
        self,
        key: str,
        value: Callable[[], tuple[Number, Number]] | tuple[Number, Number],
        *,
        namespace: str | None = None,
    ) -> None:
        """Logs a distribution value.

        Args:
            key: The key being logged
            value: The distribution value being logged, a tuple of (mean, std)
            namespace: An optional logging namespace
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def distribution_future() -> LogDistribution:
            with ContextTimer() as timer:
                mean, std = value() if callable(value) else value
            logger.debug("Distribution Key: %s, Time: %s", key, timer.elapsed_time)
            return LogDistribution(mean=mean, std=std)

        self.distributions[namespace][key] = distribution_future

    def log_histogram(
        self,
        key: str,
        value: Callable[[], np.ndarray | Array] | np.ndarray | Array,
        *,
        bins: int = 100,
        namespace: str | None = None,
    ) -> None:
        """Logs a histogram value.

        Args:
            key: The key being logged
            value: The histogram value being logged
            bins: The number of bins to use for the histogram
            namespace: An optional logging namespace
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def histogram_future() -> LogHistogram:
            with ContextTimer() as timer:
                values = value() if callable(value) else value
                values = values.reshape(-1)  # Must be flat.

                if isinstance(values, Array):
                    counts, limits = jnp.histogram(values, bins=bins)
                    counts, limits = as_numpy(counts), as_numpy(limits)
                elif isinstance(values, np.ndarray):
                    counts, limits = np.histogram(values, bins=bins)
                else:
                    raise ValueError(f"Unsupported histogram type: {type(values)}")

                histogram_values = LogHistogram(
                    min=float(values.min()),
                    max=float(values.max()),
                    num=int(values.size),
                    sum=float(values.sum()),
                    sum_squares=float(values.dot(values)),
                    bucket_limits=cast(list[float], limits[1:].tolist()),
                    bucket_counts=cast(list[int], counts.tolist()),
                )

            logger.debug("Histogram Key: %s, Time: %s", key, timer.elapsed_time)
            return histogram_values

        self.histograms[namespace][key] = histogram_future

    def log_histogram_raw(
        self,
        key: str,
        counts: Array | np.ndarray,
        limits: Array | np.ndarray,
        minv: Number | None = None,
        maxv: Number | None = None,
        sumv: Number | None = None,
        sum_squaresv: Number | None = None,
        *,
        namespace: str | None = None,
    ) -> None:
        """Logs a histogram from raw counts and limits.

        Args:
            key: The key being logged
            counts: The counts of the histogram
            limits: The limits of the histogram
            minv: The minimum value of the histogram
            maxv: The maximum value of the histogram
            sumv: The sum of the histogram
            sum_squaresv: The sum of the squares of the histogram
            namespace: An optional logging namespace
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def histogram_future() -> LogHistogram:
            with ContextTimer() as timer:
                counts_np = (as_numpy(counts) if isinstance(counts, Array) else counts).astype(int)
                limits_np = (as_numpy(limits) if isinstance(limits, Array) else limits).astype(float)

                minv_ = counts_np.min() if minv is None else minv
                maxv_ = counts_np.max() if maxv is None else maxv
                sumv_ = counts_np.sum() if sumv is None else sumv
                sum_squaresv_ = counts_np.dot(counts_np) if sum_squaresv is None else sum_squaresv

                histogram_values = LogHistogram(
                    min=float(minv_),
                    max=float(maxv_),
                    num=int(counts_np.size),
                    sum=float(sumv_),
                    sum_squares=float(sum_squaresv_),
                    bucket_limits=cast(list[float], limits_np.tolist()),
                    bucket_counts=cast(list[int], counts_np.tolist()),
                )

            logger.debug("Raw Histogram Key: %s, Time: %s", key, timer.elapsed_time)
            return histogram_values

        self.histograms[namespace][key] = histogram_future

    def log_string(
        self,
        key: str,
        value: Callable[[], str] | str,
        *,
        namespace: str | None = None,
        secondary: bool = False,
    ) -> None:
        """Logs a string value.

        Args:
            key: The key being logged
            value: The string value being logged
            namespace: An optional logging namespace
            secondary: If set, treat this as a secondary value (meaning, it is
                less important than other values, and some downstream loggers
                will not display it)
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def value_future() -> LogString:
            return LogString(value=value() if callable(value) else value, secondary=secondary)

        self.strings[namespace][key] = value_future

    def log_image(
        self,
        key: str,
        value: Callable[[], np.ndarray | Array | PILImage] | np.ndarray | Array | PILImage,
        *,
        namespace: str | None = None,
        target_resolution: tuple[int, int] | None = (512, 512),
    ) -> None:
        """Logs an image.

        Args:
            key: The key being logged
            value: The image being logged
            namespace: An optional logging namespace
            target_resolution: The target resolution for each image; if None,
                don't resample the images
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def image_future() -> LogImage:
            with ContextTimer() as timer:
                image = get_image(value() if callable(value) else value, target_resolution)
            logger.debug("Image Key: %s, Time: %s", key, timer.elapsed_time)
            return image

        self.images[namespace][key] = image_future

    def log_labeled_image(
        self,
        key: str,
        value: Callable[[], tuple[np.ndarray | Array | PILImage, str]] | tuple[np.ndarray | Array | PILImage, str],
        *,
        namespace: str | None = None,
        max_line_length: int | None = None,
        max_num_lines: int | None = None,
        target_resolution: tuple[int, int] | None = (512, 512),
        line_spacing: int = 2,
        centered: bool = True,
    ) -> None:
        """Logs an image with a label.

        Args:
            key: The key being logged
            value: The image and label being logged
            namespace: An optional logging namespace
            max_line_length: The maximum line length for the label
            max_num_lines: The number of lines of spacing to add to the bottom
                of the image
            target_resolution: The target resolution for each image; if None,
                don't resample the images
            line_spacing: The spacing between adjacent lines
            centered: If set, center the text labels, otherwise align to the left
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def image_future() -> LogImage:
            with ContextTimer() as timer:
                image, label = value() if callable(value) else value
                image = get_image(image, target_resolution)

                image_value = image_with_text(
                    image.image,
                    standardize_text(label, max_line_length),
                    max_num_lines=max_num_lines,
                    line_spacing=line_spacing,
                    centered=centered,
                )

            logger.debug("Labeled Image Key: %s, Time: %s", key, timer.elapsed_time)
            return image_value

        self.images[namespace][key] = image_future

    def log_images(
        self,
        key: str,
        value: (
            Callable[[], Sequence[np.ndarray | Array | PILImage] | np.ndarray | Array]
            | Sequence[np.ndarray | Array | PILImage]
            | np.ndarray
            | Array
        ),
        *,
        namespace: str | None = None,
        max_images: int | None = None,
        target_resolution: tuple[int, int] | None = (256, 256),
        sep: int = 0,
    ) -> None:
        """Logs a set of images.

        The images are tiled to be nearly-square.

        Args:
            key: The key being logged
            value: The images being logged
            namespace: An optional logging namespace
            max_images: The maximum number of images to show; extra images
                are clipped
            target_resolution: The target resolution for each image; if None,
                don't resample the images
            sep: An optional separation amount between adjacent images
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def images_future() -> LogImage:
            with ContextTimer() as timer:
                images = value() if callable(value) else value
                if max_images is not None:
                    images = images[:max_images]
                if isinstance(images, Array):
                    images = as_numpy(images)
                if isinstance(images, Sequence):
                    images = list(images)
                images = [get_image(image, target_resolution) for image in images]
                tiled = tile_images([img.image for img in images], sep)

            logger.debug("Images Key: %s, Time: %s", key, timer.elapsed_time)
            return LogImage(image=tiled)

        self.images[namespace][key] = images_future

    def log_labeled_images(
        self,
        key: str,
        value: (
            Callable[[], tuple[Sequence[np.ndarray | Array | PILImage] | np.ndarray | Array, Sequence[str]]]
            | tuple[Sequence[np.ndarray | Array | PILImage] | np.ndarray | Array, Sequence[str]]
        ),
        *,
        namespace: str | None = None,
        max_images: int | None = None,
        max_line_length: int | None = None,
        max_num_lines: int | None = None,
        target_resolution: tuple[int, int] | None = (256, 256),
        line_spacing: int = 2,
        centered: bool = True,
        sep: int = 0,
    ) -> None:
        """Logs a set of images with labels.

        The images are tiled to be nearly-square.

        Args:
            key: The key being logged
            value: The images and labels being logged
            namespace: An optional logging namespace
            max_images: The maximum number of images to show; extra images
                are clipped
            max_line_length: The maximum line length for the label
            max_num_lines: The number of lines of spacing to add to the bottom
                of the image
            target_resolution: The target resolution for each image; if None,
                don't resample the images
            line_spacing: The spacing between adjacent lines
            centered: If set, center the text labels, otherwise align to the left
            sep: An optional separation amount between adjacent images
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def images_future() -> LogImage:
            with ContextTimer() as timer:
                images, labels = value() if callable(value) else value
                if max_images is not None:
                    images = images[:max_images]
                    labels = labels[:max_images]
                images = [get_image(image, target_resolution) for image in images]
                labeled = [
                    image_with_text(
                        img.image,
                        standardize_text(label, max_line_length),
                        max_num_lines=max_num_lines,
                        line_spacing=line_spacing,
                        centered=centered,
                    )
                    for img, label in zip(images, labels)
                ]
                tiled = tile_images([img.image for img in labeled], sep)

            logger.debug("Labeled Images Key: %s, Time: %s", key, timer.elapsed_time)
            return LogImage(image=tiled)

        self.images[namespace][key] = images_future

    def log_file(self, name: str, contents: str) -> None:
        for logger in self.loggers:
            logger.log_file(name, contents)

    def log_video(
        self,
        key: str,
        value: Callable[[], np.ndarray | Array] | np.ndarray | Array,
        *,
        fps: int = 30,
        namespace: str | None = None,
    ) -> None:
        """Logs a video.

        Args:
            key: The key being logged
            value: The video frames. Can be:
                - A numpy array of shape (T,H,W,C) or (T,C,H,W)
                - A JAX array of shape (T,H,W,C) or (T,C,H,W)
            fps: Frames per second
            namespace: An optional logging namespace
        """
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def video_future() -> LogVideo:
            with ContextTimer() as timer:
                video = get_video(value() if callable(value) else value, fps=fps)

            logger.debug("Video Key: %s, Time: %s", key, timer.elapsed_time)
            return video

        self.videos[namespace][key] = video_future

    def log_mesh(
        self,
        key: str,
        vertices: np.ndarray | Array | Callable[[], np.ndarray | Array],
        colors: np.ndarray | Array | None | Callable[[], np.ndarray | Array | None] = None,
        faces: np.ndarray | Array | None | Callable[[], np.ndarray | Array | None] = None,
        config_dict: dict[str, Any] | None = None,
        *,
        namespace: str | None = None,
    ) -> None:
        if not self.active:
            raise RuntimeError("The logger is not active")
        namespace = self.resolve_namespace(namespace)

        @functools.lru_cache(maxsize=None)
        def mesh_future() -> LogMesh:
            with ContextTimer() as timer:
                # Get the raw values
                vertices_val = vertices() if callable(vertices) else vertices
                colors_val = colors() if callable(colors) else colors
                faces_val = faces() if callable(faces) else faces

                # Convert to numpy arrays with proper type handling
                vertices_np = as_numpy(vertices_val)
                colors_np = as_numpy_opt(colors_val)
                faces_np = as_numpy_opt(faces_val)

                # Checks vertices shape.
                if vertices_np.ndim == 2:
                    vertices_np = vertices_np[None]
                if vertices_np.shape[-1] != 3 or vertices_np.ndim != 3:
                    raise ValueError("Vertices must have shape (N, 3) or (B, N, 3)")

                # Checks colors shape.
                if colors_np is not None:
                    if colors_np.ndim == 2:
                        colors_np = colors_np[None]
                    if colors_np.shape[-1] != 3 or colors_np.ndim != 3:
                        raise ValueError("Colors must have shape (N, 3) or (B, N, 3)")

                # Checks faces shape.
                if faces_np is not None:
                    if faces_np.ndim == 2:
                        faces_np = faces_np[None]
                    if faces_np.shape[-1] != 3 or faces_np.ndim != 3:
                        raise ValueError("Faces must have shape (N, 3) or (B, N, 3)")

                # Ensures colors dtype is uint8.
                if colors_np is not None:
                    if colors_np.dtype != np.uint8:
                        colors_np = (colors_np * 255).astype(np.uint8)

                # Ensures faces dtype is int32.
                if faces_np is not None:
                    if faces_np.dtype != np.int32:
                        faces_np = faces_np.astype(np.int32)

            logger.debug("Mesh Key: %s, Time: %s", key, timer.elapsed_time)
            return LogMesh(
                vertices=vertices_np,
                colors=colors_np,
                faces=faces_np,
                config_dict=config_dict,
            )

        self.meshes[namespace][key] = mesh_future

    def __enter__(self) -> Self:
        self.active = True
        for logger in self.loggers:
            logger.start()
        return self

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        for logger in self.loggers:
            logger.stop()
        self.active = False



"""Defines utility functions for interfacing with Tensorboard."""

import functools
import io
import json
import os
import tempfile
import time
from pathlib import Path
from typing import Any, Literal, TypedDict

import numpy as np
import PIL.Image
from PIL.Image import Image as PILImage
from tensorboard.compat.proto.config_pb2 import RunMetadata
from tensorboard.compat.proto.event_pb2 import Event, TaggedRunMetadata
from tensorboard.compat.proto.graph_pb2 import GraphDef
from tensorboard.compat.proto.summary_pb2 import (
    HistogramProto,
    Summary,
    SummaryMetadata,
)
from tensorboard.compat.proto.tensor_pb2 import TensorProto
from tensorboard.compat.proto.tensor_shape_pb2 import TensorShapeProto
from tensorboard.plugins.mesh import metadata as mesh_metadata
from tensorboard.plugins.mesh.plugin_data_pb2 import MeshPluginData
from tensorboard.plugins.text.plugin_data_pb2 import TextPluginData
from tensorboard.summary.writer.event_file_writer import EventFileWriter

from xax.core.state import Phase

ImageShape = Literal["HWC", "CHW", "HW", "NHWC", "NCHW", "NHW"]


def make_histogram(values: np.ndarray, bins: str | np.ndarray, max_bins: int | None = None) -> HistogramProto:
    """Convert values into a histogram proto using logic from histogram.cc.

    Args:
        values: Input values to create histogram from
        bins: Bin specification (string like 'auto' or array of bin edges)
        max_bins: Optional maximum number of bins

    Returns:
        HistogramProto containing the histogram data
    """
    if values.size == 0:
        raise ValueError("The input has no element.")
    values = values.reshape(-1)
    counts, limits = np.histogram(values, bins=bins)
    num_bins = len(counts)

    if max_bins is not None and num_bins > max_bins:
        subsampling = num_bins // max_bins
        subsampling_remainder = num_bins % subsampling
        if subsampling_remainder != 0:
            counts = np.pad(
                counts,
                pad_width=[[0, subsampling - subsampling_remainder]],
                mode="constant",
                constant_values=0,
            )
        counts = counts.reshape(-1, subsampling).sum(axis=-1)
        new_limits = np.empty((counts.size + 1,), limits.dtype)
        new_limits[:-1] = limits[:-1:subsampling]
        new_limits[-1] = limits[-1]
        limits = new_limits

    # Find the first and the last bin defining the support of the histogram:
    cum_counts = np.cumsum(np.greater(counts, 0))
    start, end = np.searchsorted(cum_counts, [0, cum_counts[-1] - 1], side="right")
    start = int(start)
    end = int(end) + 1

    # TensorBoard only includes the right bin limits. To still have the leftmost limit
    # included, we include an empty bin left.
    counts = counts[start - 1 : end] if start > 0 else np.concatenate([[0], counts[:end]])
    limits = limits[start : end + 1]

    if counts.size == 0 or limits.size == 0:
        raise ValueError("The histogram is empty, please file a bug report.")

    sum_sq = values.dot(values)
    return HistogramProto(
        min=values.min(),
        max=values.max(),
        num=len(values),
        sum=values.sum(),
        sum_squares=sum_sq,
        bucket_limit=limits.tolist(),
        bucket=counts.tolist(),
    )


def _get_json_config(config_dict: dict[str, Any] | None) -> str:
    json_config = "{}"
    if config_dict is not None:
        json_config = json.dumps(config_dict, sort_keys=True)
    return json_config


def make_mesh_summary(
    tag: str,
    vertices: np.ndarray,
    colors: np.ndarray | None,
    faces: np.ndarray | None,
    config_dict: dict[str, Any] | None,
    display_name: str | None = None,
    description: str | None = None,
) -> Summary:
    json_config = _get_json_config(config_dict)

    summaries = []
    tensors = [
        (vertices, MeshPluginData.VERTEX),
        (faces, MeshPluginData.FACE),
        (colors, MeshPluginData.COLOR),
    ]
    # Filter out None tensors and explicitly type the list
    valid_tensors = [(t, content_type) for t, content_type in tensors if t is not None]
    components = mesh_metadata.get_components_bitmask([content_type for (_, content_type) in valid_tensors])

    for tensor, content_type in valid_tensors:  # Now we know tensor is not None
        tensor_metadata = mesh_metadata.create_summary_metadata(
            tag,
            display_name,
            content_type,
            components,
            tensor.shape,  # Safe now since tensor is not None
            description,
            json_config=json_config,
        )

        tensor_proto = TensorProto(
            dtype="DT_FLOAT",
            float_val=tensor.reshape(-1).tolist(),  # Safe now since tensor is not None
            tensor_shape=TensorShapeProto(
                dim=[
                    TensorShapeProto.Dim(size=tensor.shape[0]),  # Safe now since tensor is not None
                    TensorShapeProto.Dim(size=tensor.shape[1]),
                    TensorShapeProto.Dim(size=tensor.shape[2]),
                ]
            ),
        )

        tensor_summary = Summary.Value(
            tag=mesh_metadata.get_instance_name(tag, content_type),
            tensor=tensor_proto,
            metadata=tensor_metadata,
        )

        summaries.append(tensor_summary)

    return Summary(value=summaries)


class TensorboardProtobufWriter:
    def __init__(
        self,
        log_directory: str | Path,
        max_queue_size: int = 10,
        flush_seconds: float = 120.0,
        filename_suffix: str = "",
    ) -> None:
        super().__init__()

        self.log_directory = Path(log_directory)
        self.max_queue_size = max_queue_size
        self.flush_seconds = flush_seconds
        self.filename_suffix = filename_suffix

    @functools.cached_property
    def event_writer(self) -> EventFileWriter:
        return EventFileWriter(
            logdir=str(self.log_directory),
            max_queue_size=self.max_queue_size,
            flush_secs=self.flush_seconds,
            filename_suffix=self.filename_suffix,
        )

    def add_event(
        self,
        event: Event,
        step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        event.wall_time = time.time() if walltime is None else walltime
        if step is not None:
            event.step = int(step)
        self.event_writer.add_event(event)

    def add_summary(
        self,
        summary: Summary,
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        event = Event(summary=summary)
        self.add_event(event, step=global_step, walltime=walltime)

    def add_graph(
        self,
        graph: GraphDef,
        run_metadata: RunMetadata | None = None,
        walltime: float | None = None,
    ) -> None:
        event = Event(graph_def=graph.SerializeToString())
        self.add_event(event, walltime=walltime)
        if run_metadata is not None:
            trm = TaggedRunMetadata(tag="step1", run_metadata=run_metadata.SerializeToString())
            event = Event(tagged_run_metadata=trm)
            self.add_event(event, walltime=walltime)

    def flush(self) -> None:
        self.event_writer.flush()

    def close(self) -> None:
        self.event_writer.close()


class TensorboardWriter:
    """Defines a class for writing artifacts to Tensorboard.

    Parameters:
        log_directory: The directory to write logs to.
        max_queue_size: The maximum queue size.
        flush_seconds: How often to flush logs.
        filename_suffix: The filename suffix to use.
    """

    def __init__(
        self,
        log_directory: str | Path,
        max_queue_size: int = 10,
        flush_seconds: float = 120.0,
        filename_suffix: str = "",
    ) -> None:
        super().__init__()

        self.pb_writer = TensorboardProtobufWriter(
            log_directory=log_directory,
            max_queue_size=max_queue_size,
            flush_seconds=flush_seconds,
            filename_suffix=filename_suffix,
        )

    def add_scalar(
        self,
        tag: str,
        value: float,
        global_step: int | None = None,
        walltime: float | None = None,
        new_style: bool = True,
        double_precision: bool = False,
    ) -> None:
        if new_style:
            self.pb_writer.add_summary(
                Summary(
                    value=[
                        Summary.Value(
                            tag=tag,
                            tensor=(
                                TensorProto(double_val=[value], dtype="DT_DOUBLE")
                                if double_precision
                                else TensorProto(float_val=[value], dtype="DT_FLOAT")
                            ),
                            metadata=SummaryMetadata(
                                plugin_data=SummaryMetadata.PluginData(
                                    plugin_name="scalars",
                                ),
                            ),
                        )
                    ],
                ),
                global_step=global_step,
                walltime=walltime,
            )
        else:
            self.pb_writer.add_summary(
                Summary(
                    value=[
                        Summary.Value(
                            tag=tag,
                            simple_value=value,
                        ),
                    ],
                ),
                global_step=global_step,
                walltime=walltime,
            )

    def add_image(
        self,
        tag: str,
        value: PILImage,
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        output = io.BytesIO()
        value.convert("RGB").save(output, format="PNG")
        image_string = output.getvalue()
        output.close()

        self.pb_writer.add_summary(
            Summary(
                value=[
                    Summary.Value(
                        tag=tag,
                        image=Summary.Image(
                            height=value.height,
                            width=value.width,
                            colorspace=3,  # RGB
                            encoded_image_string=image_string,
                        ),
                    ),
                ],
            ),
            global_step=global_step,
            walltime=walltime,
        )

    def add_video(
        self,
        tag: str,
        value: np.ndarray,
        global_step: int | None = None,
        walltime: float | None = None,
        fps: int = 30,
    ) -> None:
        assert value.ndim == 4, "Video must be 4D array (T, H, W, C)"

        images = [PIL.Image.fromarray(frame).convert("RGB") for frame in value]
        width, height = images[0].size
        big_image = PIL.Image.new("RGB", (width, height * len(images)))
        for i, im in enumerate(images):
            big_image.paste(im, (0, i * height))

        quantized_big = big_image.quantize(method=PIL.Image.Quantize.MAXCOVERAGE, dither=PIL.Image.Dither.NONE)
        palette = quantized_big.getpalette()

        processed = []
        for im in images:
            q = im.quantize(
                method=PIL.Image.Quantize.MAXCOVERAGE,
                palette=quantized_big,
                dither=PIL.Image.Dither.NONE,
            )
            processed.append(q)

        if palette is not None:
            palette[0:3] = [255, 255, 255]
            for im in processed:
                im.putpalette(palette)

        # Create temporary file for GIF
        temp_file = tempfile.NamedTemporaryFile(suffix=".gif", delete=False)
        try:
            processed[0].save(
                temp_file.name,
                save_all=True,
                append_images=processed[1:],
                duration=int(1000 / fps),
                loop=0,
            )
            with open(temp_file.name, "rb") as f:
                video_string = f.read()

        finally:
            # Clean up temporary file
            try:
                os.remove(temp_file.name)
            except OSError:
                pass

        # Add to summary
        self.pb_writer.add_summary(
            Summary(
                value=[
                    Summary.Value(
                        tag=tag,
                        image=Summary.Image(
                            height=value.shape[1],
                            width=value.shape[2],
                            colorspace=value.shape[3],
                            encoded_image_string=video_string,
                        ),
                    ),
                ],
            ),
            global_step=global_step,
            walltime=walltime,
        )

    def add_text(
        self,
        tag: str,
        value: str,
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        self.pb_writer.add_summary(
            Summary(
                value=[
                    Summary.Value(
                        tag=tag + "/text_summary",
                        metadata=SummaryMetadata(
                            plugin_data=SummaryMetadata.PluginData(
                                plugin_name="text", content=TextPluginData(version=0).SerializeToString()
                            ),
                        ),
                        tensor=TensorProto(
                            dtype="DT_STRING",
                            string_val=[value.encode(encoding="utf_8")],
                            tensor_shape=TensorShapeProto(dim=[TensorShapeProto.Dim(size=1)]),
                        ),
                    ),
                ],
            ),
            global_step=global_step,
            walltime=walltime,
        )

    def add_histogram(
        self,
        tag: str,
        values: np.ndarray,
        global_step: int | None = None,
        bins: str | np.ndarray = "auto",
        walltime: float | None = None,
        max_bins: int | None = None,
    ) -> None:
        hist = make_histogram(values.astype(float), bins, max_bins)
        self.pb_writer.add_summary(
            Summary(value=[Summary.Value(tag=tag, histo=hist)]),
            global_step=global_step,
            walltime=walltime,
        )

    def add_histogram_raw(
        self,
        tag: str,
        min: float | np.ndarray,
        max: float | np.ndarray,
        num: int | np.ndarray,
        sum: float | np.ndarray,
        sum_squares: float | np.ndarray,
        bucket_limits: list[float | np.ndarray],
        bucket_counts: list[int | np.ndarray],
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        """Add histogram with raw data to summary.

        Args:
            tag: Data identifier
            min: Min value
            max: Max value
            num: Number of values
            sum: Sum of all values
            sum_squares: Sum of squares for all values
            bucket_limits: Upper value per bucket
            bucket_counts: Number of values per bucket
            global_step: Global step value to record
            walltime: Optional override default walltime
        """
        if len(bucket_limits) != len(bucket_counts):
            raise ValueError("len(bucket_limits) != len(bucket_counts)")

        # Convert numpy arrays to Python types
        hist = HistogramProto(
            min=float(min),
            max=float(max),
            num=int(num),
            sum=float(sum),
            sum_squares=float(sum_squares),
            bucket_limit=[float(x) for x in bucket_limits],
            bucket=[int(x) for x in bucket_counts],
        )
        self.pb_writer.add_summary(
            Summary(value=[Summary.Value(tag=tag, histo=hist)]),
            global_step=global_step,
            walltime=walltime,
        )

    def add_gaussian_distribution(
        self,
        tag: str,
        mean: float | np.ndarray,
        std: float | np.ndarray,
        bins: int = 16,
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        """Add a Gaussian distribution to the summary.

        Args:
            tag: Data identifier
            mean: Mean of the Gaussian distribution
            std: Standard deviation of the Gaussian distribution
            bins: Number of bins to use for the histogram
            global_step: Global step value to record
            walltime: Optional override default walltime
        """
        # Convert numpy arrays to Python types
        mean = float(mean)
        std = float(std)

        # Create bin edges spanning ±3 standard deviations
        bin_edges = np.linspace(mean - 3 * std, mean + 3 * std, bins + 1)

        # Calculate the probability density for each bin
        bin_centers = (bin_edges[1:] + bin_edges[:-1]) / 2
        gaussian_pdf = np.exp(-0.5 * ((bin_centers - mean) / std) ** 2) / (std * np.sqrt(2 * np.pi))

        # Scale the PDF to represent counts
        num_samples = bins * 1000
        bucket_counts = (gaussian_pdf * num_samples * (bin_edges[1] - bin_edges[0])).astype(int)

        # Ensure we have at least one count per bin for visualization
        bucket_counts = np.maximum(bucket_counts, 1)

        # Calculate actual statistics based on the discretized distribution
        total_counts = float(bucket_counts.sum())
        weighted_sum = float((bin_centers * bucket_counts).sum())
        weighted_sum_squares = float((bin_centers**2 * bucket_counts).sum())

        # Convert bin edges to list of floats explicitly
        bucket_limits: list[float | np.ndarray] = [float(x) for x in bin_edges[1:]]

        self.add_histogram_raw(
            tag=tag,
            min=float(bin_edges[0]),
            max=float(bin_edges[-1]),
            num=int(total_counts),
            sum=weighted_sum,
            sum_squares=weighted_sum_squares,
            bucket_limits=bucket_limits,  # Now properly typed
            bucket_counts=bucket_counts.tolist(),
            global_step=global_step,
            walltime=walltime,
        )

    def add_mesh(
        self,
        tag: str,
        vertices: np.ndarray,
        colors: np.ndarray | None,
        faces: np.ndarray | None,
        config_dict: dict[str, Any] | None,
        global_step: int | None = None,
        walltime: float | None = None,
    ) -> None:
        self.pb_writer.add_summary(
            make_mesh_summary(tag, vertices, colors, faces, config_dict),
            global_step=global_step,
            walltime=walltime,
        )


class TensorboardWriterKwargs(TypedDict):
    max_queue_size: int
    flush_seconds: float
    filename_suffix: str


class TensorboardWriters:
    def __init__(
        self,
        log_directory: str | Path,
        max_queue_size: int = 10,
        flush_seconds: float = 120.0,
        filename_suffix: str = "",
    ) -> None:
        super().__init__()

        self.log_directory = Path(log_directory)

        self.kwargs: TensorboardWriterKwargs = {
            "max_queue_size": max_queue_size,
            "flush_seconds": flush_seconds,
            "filename_suffix": filename_suffix,
        }

    @functools.cached_property
    def train_writer(self) -> TensorboardWriter:
        return TensorboardWriter(self.log_directory / "train", **self.kwargs)

    @functools.cached_property
    def valid_writer(self) -> TensorboardWriter:
        return TensorboardWriter(self.log_directory / "valid", **self.kwargs)

    def writer(self, phase: Phase) -> TensorboardWriter:
        match phase:
            case "train":
                return self.train_writer
            case "valid":
                return self.valid_writer
            case _:
                raise NotImplementedError(f"Unexpected phase: {phase}")



"""Export JAX functions to TensorFlow SavedModel format."""

import logging
from pathlib import Path
from typing import Callable

import jax
from jaxtyping import Array, PyTree

try:
    import flax
    import tensorflow as tf
    from jax.experimental import jax2tf
    from orbax.export import ExportManager, JaxModule, ServingConfig
except ImportError as e:
    raise ImportError(
        "In order to export models, please install Xax with exportable dependencies, "
        "using 'xax[exportable]` to install the required dependencies."
    ) from e

logger = logging.getLogger(__name__)


def _run_infer(tf_module: tf.Module, input_shapes: list[tuple[int, ...]], batch_size: int | None) -> tf.Tensor:
    """Warm up the model by running it once."""
    if batch_size is not None:
        test_inputs = [
            jax.random.normal(jax.random.PRNGKey(42), (batch_size, *input_shape)) for input_shape in input_shapes
        ]
    else:
        test_inputs = [jax.random.normal(jax.random.PRNGKey(42), (1, *input_shape)) for input_shape in input_shapes]
    if not hasattr(tf_module, "infer"):
        raise ValueError("Model does not have an infer method")
    return tf_module.infer(*test_inputs)


def export(
    model: Callable,
    input_shapes: list[tuple[int, ...]],
    output_dir: str | Path = "export",
    batch_size: int | None = None,
) -> None:
    """Export a JAX function to TensorFlow SavedModel.

    Note: Tensorflow GraphDef can't be larger than 2GB - https://github.com/tensorflow/tensorflow/issues/51870
    You can avoid this by saving model parameters as non-constants.

    Args:
        model: The JAX function to export.
        input_shapes: The shape of the input tensors, excluding batch dimension.
        output_dir: Directory to save the exported model.
        batch_size: Optional batch dimension. If None, a polymorphic batch dimension is used.
    """
    tf_module = tf.Module()
    # Create a polymorphic shape specification for each input
    poly_spec = "(b, ...)" if batch_size is not None else "(None, ...)"
    polymorphic_shapes = [poly_spec] * len(input_shapes)
    tf_module.infer = tf.function(  # type: ignore [attr-defined]
        jax2tf.convert(
            model,
            polymorphic_shapes=polymorphic_shapes,
            # setting this to False will allow the model to run on platforms other than the one that exports the model
            # https://github.com/jax-ml/jax/blob/051687dc4c899df3d95c30b812ade401d8b31166/jax/experimental/jax2tf/README.md?plain=1#L1342
            # generally though I think native_serialization is recommended
            native_serialization=False,
            with_gradient=False,
        ),
        autograph=False,
        input_signature=[tf.TensorSpec([batch_size] + list(input_shape), tf.float32) for input_shape in input_shapes],
    )

    # warm up the model
    _run_infer(tf_module, input_shapes, batch_size)

    logger.info("Exporting SavedModel to %s", output_dir)
    tf.saved_model.save(
        tf_module,
        output_dir,
    )


def export_with_params(
    model: Callable,
    params: PyTree,
    input_shapes: list[tuple[int, ...]],
    output_dir: str | Path = "export",
    batch_dim: int | None = None,
) -> None:
    """Export a JAX function that takes parameters to TensorFlow SavedModel.

    Args:
        model: The JAX function to export. Should take parameters as first argument.
        params: The parameters to use for the model.
        input_shapes: The shape of the input tensors, excluding batch dimension.
        output_dir: Directory to save the exported model.
        batch_dim: Optional batch dimension. If None, a polymorphic batch dimension is used.
    """
    param_vars = tf.nest.map_structure(tf.Variable, params)

    converted_model = jax2tf.convert(model)

    def model_fn(*inputs: PyTree) -> Array:
        return converted_model(param_vars, *inputs)

    tf_module = tf.Module()
    tf_module._variables = tf.nest.flatten(param_vars)  # type: ignore [attr-defined]
    tf_module.infer = tf.function(  # type: ignore [attr-defined]
        model_fn,
        jit_compile=True,
        autograph=False,
        input_signature=[tf.TensorSpec([batch_dim] + list(input_shape), tf.float32) for input_shape in input_shapes],
    )

    # warm up the model
    _run_infer(tf_module, input_shapes, batch_dim)

    logger.info("Exporting SavedModel to %s", output_dir)
    tf.saved_model.save(tf_module, output_dir)


def export_flax(
    model: flax.linen.Module,
    params: PyTree,
    input_shape: tuple[int, ...],
    preprocessor: Callable | None = None,
    postprocessor: Callable | None = None,
    input_name: str = "inputs",
    output_name: str = "outputs",
    output_dir: str | Path = "export",
) -> None:
    jax_module = JaxModule(
        params, model.apply, trainable=False, input_polymorphic_shape="(b, ...)"
    )  # if you want to use a batch dimension

    # to avoid mapping sequences to ambiguous mappings
    if postprocessor is None:

        def postprocessor(x: PyTree) -> PyTree:
            return {output_name: x}

    export_manager = ExportManager(
        jax_module,
        [
            ServingConfig(
                "serving_default",
                input_signature=[tf.TensorSpec([None] + list(input_shape), tf.float32, name=input_name)],
                tf_preprocessor=preprocessor,
                tf_postprocessor=postprocessor,
            )
        ],
    )

    logger.info("Exporting model to %s", output_dir)
    export_manager.save(output_dir)



"""Defines a mixin to wrap some steps in a context manager."""

import time
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, ContextManager, TypeVar

import jax

from xax.task.base import BaseConfig, BaseTask


class StepContext(ContextManager):
    """Context manager to get the current step type."""

    CURRENT_STEP: str | None = None

    def __init__(
        self,
        step: str,
        on_context_start: Callable[[str], None],
        on_context_end: Callable[[str, float], None],
    ) -> None:
        self.step = step
        self.start_time = 0.0
        self.on_context_start = on_context_start
        self.on_context_end = on_context_end

    def __enter__(self) -> None:
        StepContext.CURRENT_STEP = self.step
        self.start_time = time.time()
        self.on_context_start(self.step)

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        StepContext.CURRENT_STEP = None
        self.on_context_end(self.step, time.time() - self.start_time)


@jax.tree_util.register_dataclass
@dataclass
class StepContextConfig(BaseConfig):
    pass


Config = TypeVar("Config", bound=StepContextConfig)


class StepContextMixin(BaseTask[Config]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)

    def step_context(self, step: str) -> ContextManager:
        return StepContext(step, self.on_context_start, self.on_context_stop)

    def on_context_start(self, step: str) -> None:
        pass

    def on_context_stop(self, step: str, elapsed_time: float) -> None:
        pass



"""Defines a logger that calls a callback function with the log line."""

from typing import Callable

from xax.task.logger import LogError, LogErrorSummary, LoggerImpl, LogLine, LogPing, LogStatus


class CallbackLogger(LoggerImpl):
    def __init__(
        self,
        *,
        callback: Callable[[LogLine], None] = lambda x: None,
        error_summary_callback: Callable[[LogErrorSummary], None] = lambda x: None,
        error_callback: Callable[[LogError], None] = lambda x: None,
        status_callback: Callable[[LogStatus], None] = lambda x: None,
        ping_callback: Callable[[LogPing], None] = lambda x: None,
        file_callback: Callable[[str, str], None] = lambda x, y: None,
    ) -> None:
        super().__init__()

        self.callback = callback
        self.error_summary_callback = error_summary_callback
        self.error_callback = error_callback
        self.status_callback = status_callback
        self.ping_callback = ping_callback
        self.file_callback = file_callback

    def write(self, line: LogLine) -> None:
        self.callback(line)

    def write_error_summary(self, error_summary: LogErrorSummary) -> None:
        self.error_summary_callback(error_summary)

    def write_error(self, error: LogError) -> None:
        self.error_callback(error)

    def write_status(self, status: LogStatus) -> None:
        self.status_callback(status)

    def write_ping(self, ping: LogPing) -> None:
        self.ping_callback(ping)

    def log_file(self, name: str, contents: str) -> None:
        self.file_callback(name, contents)
